/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.connector.generator.swagger;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.node.ObjectNode;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.swagger.models.Swagger;
import io.syndesis.common.model.Violation;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Generated;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link SwaggerModelInfo}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new SwaggerModelInfo.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableSwaggerModelInfo.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@Generated({"Immutables.generator", "SwaggerModelInfo"})
@Immutable
final class ImmutableSwaggerModelInfo
    implements SwaggerModelInfo {
  private final List<Violation> errors;
  private final Swagger model;
  private final String resolvedSpecification;
  private final List<Violation> warnings;

  private ImmutableSwaggerModelInfo(
      Iterable<? extends Violation> errors,
      Swagger model,
      String resolvedSpecification,
      Iterable<? extends Violation> warnings) {
    this.errors = createUnmodifiableList(false, createSafeList(errors, true, false));
    this.model = model;
    this.resolvedSpecification = resolvedSpecification;
    this.warnings = createUnmodifiableList(false, createSafeList(warnings, true, false));
    initShim.errors(this.errors);
    initShim.warnings(this.warnings);
    this.initShim = null;
  }

  private ImmutableSwaggerModelInfo(ImmutableSwaggerModelInfo.Builder builder) {
    this.model = builder.model;
    this.resolvedSpecification = builder.resolvedSpecification;
    if (builder.errorsIsSet()) {
      initShim.errors(createUnmodifiableList(true, builder.errors));
    }
    if (builder.warningsIsSet()) {
      initShim.warnings(createUnmodifiableList(true, builder.warnings));
    }
    this.errors = initShim.getErrors();
    this.warnings = initShim.getWarnings();
    this.initShim = null;
  }

  private ImmutableSwaggerModelInfo(
      ImmutableSwaggerModelInfo original,
      List<Violation> errors,
      Swagger model,
      String resolvedSpecification,
      List<Violation> warnings) {
    this.errors = errors;
    this.model = model;
    this.resolvedSpecification = resolvedSpecification;
    this.warnings = warnings;
    this.initShim = null;
  }

  private static final int STAGE_INITIALIZING = -1;
  private static final int STAGE_UNINITIALIZED = 0;
  private static final int STAGE_INITIALIZED = 1;
  private transient volatile InitShim initShim = new InitShim();

  private final class InitShim {
    private List<Violation> errors;
    private int errorsBuildStage;

    List<Violation> getErrors() {
      if (errorsBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (errorsBuildStage == STAGE_UNINITIALIZED) {
        errorsBuildStage = STAGE_INITIALIZING;
        this.errors = createUnmodifiableList(false, createSafeList(getErrorsInitialize(), true, false));
        errorsBuildStage = STAGE_INITIALIZED;
      }
      return this.errors;
    }

    void errors(List<Violation> errors) {
      this.errors = errors;
      errorsBuildStage = STAGE_INITIALIZED;
    }
    private List<Violation> warnings;
    private int warningsBuildStage;

    List<Violation> getWarnings() {
      if (warningsBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (warningsBuildStage == STAGE_UNINITIALIZED) {
        warningsBuildStage = STAGE_INITIALIZING;
        this.warnings = createUnmodifiableList(false, createSafeList(getWarningsInitialize(), true, false));
        warningsBuildStage = STAGE_INITIALIZED;
      }
      return this.warnings;
    }

    void warnings(List<Violation> warnings) {
      this.warnings = warnings;
      warningsBuildStage = STAGE_INITIALIZED;
    }

    private String formatInitCycleMessage() {
      ArrayList<String> attributes = new ArrayList<String>();
      if (errorsBuildStage == STAGE_INITIALIZING) attributes.add("errors");
      if (warningsBuildStage == STAGE_INITIALIZING) attributes.add("warnings");
      return "Cannot build SwaggerModelInfo, attribute initializers form cycle" + attributes;
    }
  }

  private List<Violation> getErrorsInitialize() {
    return SwaggerModelInfo.super.getErrors();
  }

  private List<Violation> getWarningsInitialize() {
    return SwaggerModelInfo.super.getWarnings();
  }

  /**
   * @return The value of the {@code errors} attribute
   */
  @JsonProperty("errors")
  @Override
  public List<Violation> getErrors() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getErrors()
        : this.errors;
  }

  /**
   * @return The value of the {@code model} attribute
   */
  @JsonProperty("model")
  @Override
  public Swagger getModel() {
    return model;
  }

  /**
   * @return The value of the {@code resolvedSpecification} attribute
   */
  @JsonProperty("resolvedSpecification")
  @Override
  public String getResolvedSpecification() {
    return resolvedSpecification;
  }

  /**
   * @return The value of the {@code warnings} attribute
   */
  @JsonProperty("warnings")
  @Override
  public List<Violation> getWarnings() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getWarnings()
        : this.warnings;
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link SwaggerModelInfo#getErrors() errors}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSwaggerModelInfo withErrors(Violation... elements) {
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableSwaggerModelInfo(this, newValue, this.model, this.resolvedSpecification, this.warnings));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link SwaggerModelInfo#getErrors() errors}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of errors elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSwaggerModelInfo withErrors(Iterable<? extends Violation> elements) {
    if (this.errors == elements) return this;
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableSwaggerModelInfo(this, newValue, this.model, this.resolvedSpecification, this.warnings));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link SwaggerModelInfo#getModel() model} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for model (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableSwaggerModelInfo withModel(Swagger value) {
    if (this.model == value) return this;
    return validate(new ImmutableSwaggerModelInfo(this, this.errors, value, this.resolvedSpecification, this.warnings));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link SwaggerModelInfo#getResolvedSpecification() resolvedSpecification} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for resolvedSpecification (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableSwaggerModelInfo withResolvedSpecification(String value) {
    if (Objects.equals(this.resolvedSpecification, value)) return this;
    return validate(new ImmutableSwaggerModelInfo(this, this.errors, this.model, value, this.warnings));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link SwaggerModelInfo#getWarnings() warnings}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSwaggerModelInfo withWarnings(Violation... elements) {
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableSwaggerModelInfo(this, this.errors, this.model, this.resolvedSpecification, newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link SwaggerModelInfo#getWarnings() warnings}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of warnings elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSwaggerModelInfo withWarnings(Iterable<? extends Violation> elements) {
    if (this.warnings == elements) return this;
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableSwaggerModelInfo(this, this.errors, this.model, this.resolvedSpecification, newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableSwaggerModelInfo} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableSwaggerModelInfo
        && equalTo((ImmutableSwaggerModelInfo) another);
  }

  private boolean equalTo(ImmutableSwaggerModelInfo another) {
    return errors.equals(another.errors)
        && Objects.equals(model, another.model)
        && Objects.equals(resolvedSpecification, another.resolvedSpecification)
        && warnings.equals(another.warnings);
  }

  /**
   * Computes a hash code from attributes: {@code errors}, {@code model}, {@code resolvedSpecification}, {@code warnings}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + errors.hashCode();
    h += (h << 5) + Objects.hashCode(model);
    h += (h << 5) + Objects.hashCode(resolvedSpecification);
    h += (h << 5) + warnings.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code SwaggerModelInfo} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "SwaggerModelInfo{"
        + "errors=" + errors
        + ", model=" + model
        + ", resolvedSpecification=" + resolvedSpecification
        + ", warnings=" + warnings
        + "}";
  }

  private volatile long lazyInitBitmap;

  private static final long RESOLVED_JSON_GRAPH_LAZY_INIT_BIT = 0x1L;

  private ObjectNode resolvedJsonGraph;

  /**
   * {@inheritDoc}
   * <p>
   * Returns a lazily initialized value of the {@link SwaggerModelInfo#getResolvedJsonGraph() resolvedJsonGraph} attribute.
   * Initialized once and only once and stored for subsequent access with proper synchronization.
   * @return A lazily initialized value of the {@code l.name} attribute
   */
  @Override
  public ObjectNode getResolvedJsonGraph() {
    if ((lazyInitBitmap & RESOLVED_JSON_GRAPH_LAZY_INIT_BIT) == 0) {
      synchronized (this) {
        if ((lazyInitBitmap & RESOLVED_JSON_GRAPH_LAZY_INIT_BIT) == 0) {
          this.resolvedJsonGraph = Objects.requireNonNull(SwaggerModelInfo.super.getResolvedJsonGraph(), "resolvedJsonGraph");
          lazyInitBitmap |= RESOLVED_JSON_GRAPH_LAZY_INIT_BIT;
        }
      }
    }
    return resolvedJsonGraph;
  }

  /**
   * Construct a new immutable {@code SwaggerModelInfo} instance.
   * @param errors The value for the {@code errors} attribute
   * @param model The value for the {@code model} attribute
   * @param resolvedSpecification The value for the {@code resolvedSpecification} attribute
   * @param warnings The value for the {@code warnings} attribute
   * @return An immutable SwaggerModelInfo instance
   */
  public static SwaggerModelInfo of(List<Violation> errors, Swagger model, String resolvedSpecification, List<Violation> warnings) {
    return of((Iterable<? extends Violation>) errors, model, resolvedSpecification, (Iterable<? extends Violation>) warnings);
  }

  /**
   * Construct a new immutable {@code SwaggerModelInfo} instance.
   * @param errors The value for the {@code errors} attribute
   * @param model The value for the {@code model} attribute
   * @param resolvedSpecification The value for the {@code resolvedSpecification} attribute
   * @param warnings The value for the {@code warnings} attribute
   * @return An immutable SwaggerModelInfo instance
   */
  public static SwaggerModelInfo of(Iterable<? extends Violation> errors, Swagger model, String resolvedSpecification, Iterable<? extends Violation> warnings) {
    return validate(new ImmutableSwaggerModelInfo(errors, model, resolvedSpecification, warnings));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableSwaggerModelInfo validate(ImmutableSwaggerModelInfo instance) {
    Set<ConstraintViolation<ImmutableSwaggerModelInfo>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link SwaggerModelInfo} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable SwaggerModelInfo instance
   */
  public static SwaggerModelInfo copyOf(SwaggerModelInfo instance) {
    if (instance instanceof ImmutableSwaggerModelInfo) {
      return (ImmutableSwaggerModelInfo) instance;
    }
    return new SwaggerModelInfo.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link SwaggerModelInfo SwaggerModelInfo}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @NotThreadSafe
  public static class Builder {
    private static final long OPT_BIT_ERRORS = 0x1L;
    private static final long OPT_BIT_WARNINGS = 0x2L;
    private long optBits;

    private List<Violation> errors = new ArrayList<Violation>();
    private @Nullable Swagger model;
    private @Nullable String resolvedSpecification;
    private List<Violation> warnings = new ArrayList<Violation>();

    /**
     * Creates a builder for {@link SwaggerModelInfo SwaggerModelInfo} instances.
     */
    public Builder() {
      if (!(this instanceof SwaggerModelInfo.Builder)) {
        throw new UnsupportedOperationException("Use: new SwaggerModelInfo.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code SwaggerModelInfo} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder createFrom(SwaggerModelInfo instance) {
      Objects.requireNonNull(instance, "instance");
      addAllErrors(instance.getErrors());
      Swagger modelValue = instance.getModel();
      if (modelValue != null) {
        model(modelValue);
      }
      String resolvedSpecificationValue = instance.getResolvedSpecification();
      if (resolvedSpecificationValue != null) {
        resolvedSpecification(resolvedSpecificationValue);
      }
      addAllWarnings(instance.getWarnings());
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Adds one element to {@link SwaggerModelInfo#getErrors() errors} list.
     * @param element A errors element
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addError(Violation element) {
      this.errors.add(Objects.requireNonNull(element, "errors element"));
      optBits |= OPT_BIT_ERRORS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Adds elements to {@link SwaggerModelInfo#getErrors() errors} list.
     * @param elements An array of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addError(Violation... elements) {
      for (Violation element : elements) {
        this.errors.add(Objects.requireNonNull(element, "errors element"));
      }
      optBits |= OPT_BIT_ERRORS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link SwaggerModelInfo#getErrors() errors} list.
     * @param elements An iterable of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("errors")
    public final SwaggerModelInfo.Builder errors(Iterable<? extends Violation> elements) {
      this.errors.clear();
      return addAllErrors(elements);
    }

    /**
     * Adds elements to {@link SwaggerModelInfo#getErrors() errors} list.
     * @param elements An iterable of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addAllErrors(Iterable<? extends Violation> elements) {
      for (Violation element : elements) {
        this.errors.add(Objects.requireNonNull(element, "errors element"));
      }
      optBits |= OPT_BIT_ERRORS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Initializes the value for the {@link SwaggerModelInfo#getModel() model} attribute.
     * @param model The value for model (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("model")
    public final SwaggerModelInfo.Builder model(Swagger model) {
      this.model = model;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Initializes the value for the {@link SwaggerModelInfo#getResolvedSpecification() resolvedSpecification} attribute.
     * @param resolvedSpecification The value for resolvedSpecification (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("resolvedSpecification")
    public final SwaggerModelInfo.Builder resolvedSpecification(String resolvedSpecification) {
      this.resolvedSpecification = resolvedSpecification;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Adds one element to {@link SwaggerModelInfo#getWarnings() warnings} list.
     * @param element A warnings element
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addWarning(Violation element) {
      this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      optBits |= OPT_BIT_WARNINGS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Adds elements to {@link SwaggerModelInfo#getWarnings() warnings} list.
     * @param elements An array of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addWarning(Violation... elements) {
      for (Violation element : elements) {
        this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      }
      optBits |= OPT_BIT_WARNINGS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link SwaggerModelInfo#getWarnings() warnings} list.
     * @param elements An iterable of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("warnings")
    public final SwaggerModelInfo.Builder warnings(Iterable<? extends Violation> elements) {
      this.warnings.clear();
      return addAllWarnings(elements);
    }

    /**
     * Adds elements to {@link SwaggerModelInfo#getWarnings() warnings} list.
     * @param elements An iterable of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final SwaggerModelInfo.Builder addAllWarnings(Iterable<? extends Violation> elements) {
      for (Violation element : elements) {
        this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      }
      optBits |= OPT_BIT_WARNINGS;
      return (SwaggerModelInfo.Builder) this;
    }

    /**
     * Builds a new {@link SwaggerModelInfo SwaggerModelInfo}.
     * @return An immutable instance of SwaggerModelInfo
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public SwaggerModelInfo build() {
      return ImmutableSwaggerModelInfo.validate(new ImmutableSwaggerModelInfo(this));
    }

    private boolean errorsIsSet() {
      return (optBits & OPT_BIT_ERRORS) != 0;
    }

    private boolean warningsIsSet() {
      return (optBits & OPT_BIT_WARNINGS) != 0;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
